package brownstrategy.ensembles;

import game.Game;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Random;
import java.util.StringTokenizer;

import brownstrategy.agent.Agent;
import brownstrategy.agent.NondeterministicAgent;
import brownstrategy.oracles.PrevSeqIndividualWithUtilOracle;
import brownstrategy.oracles.PrevSeqOpponentsWithUtilOracle;
import brownstrategy.oracles.PrevSequenceAllOracle;
import brownstrategy.oracles.PrevSequenceIndividualOracle;
import brownstrategy.oracles.PrevSequenceAllRotSymOracle;
import brownstrategy.oracles.PrevSequenceOppSymOracle;
import brownstrategy.oracles.PrevSequenceOpponentsOracle;
import brownstrategy.oracles.SameAsLastTimeOracle;

/**
 *
 *	This is an ensemble class of oracles
 * 
 * @author jberg
 *
 */

public class EnsembleOracle extends AbstractEnsemble {

	double _numRightWeight,_probWeightWeight;
	int _numPastRounds,_numOracles;
	HashMap<Agent,ArrayList<Double>> probWeightMap;
	HashMap<Agent,ArrayList<Double>> rightMap;

	public EnsembleOracle(Game g, long seed, String options) {
		super(g,seed,options);
		_seed = seed;
		StringTokenizer st = new StringTokenizer(options, ",");
		if (st.countTokens() >= 5) {
			st.nextToken();
			_numRightWeight = Double.parseDouble(st.nextToken());
			_probWeightWeight = Double.parseDouble(st.nextToken());
			_numPastRounds = Integer.parseInt(st.nextToken());
			_numOracles = Integer.parseInt(st.nextToken());
		}
		else {
			_numRightWeight = 1.0;
			_probWeightWeight = 1.0;
			_numPastRounds = 25;
			_numOracles = 5;
		}

		probWeightMap = new HashMap<Agent,ArrayList<Double>>();
		rightMap = new HashMap<Agent,ArrayList<Double>>();
	}

	@Override
	public void addModel(Agent model) {
		super.addModel(model);
		ArrayList<Double> weightList = new ArrayList<Double>();
		ArrayList<Double> rightList = new ArrayList<Double>();
		probWeightMap.put(model,weightList);
		rightMap.put(model,rightList);
	}
	
	@Override
	public void removeModel(Agent model) {
		super.removeModel(model);
		probWeightMap.remove(model);
		rightMap.remove(model);
	}

	@Override
	public void observeOutcome(int[] actions) {
		/*
		 * Rank all of the oracles
		 */
		for(Agent agent : _lastPredictions.keySet()) {
			ArrayList<Double> probWeightList = probWeightMap.get(agent);
			ArrayList<Double> rightList = rightMap.get(agent);
			double[] lastPrediction = _lastPredictions.get(agent);
			if(lastPrediction == null) {
				continue;
			}
			probWeightList.add(lastPrediction[actions[_playerIdx]]);
			rightList.add(1.0*isMax(lastPrediction[actions[_playerIdx]],lastPrediction));
			probWeightMap.put(agent, probWeightList);
			rightMap.put(agent, rightList);
		}
		
		/*
		 * Derive ensemble weights
		 */
		ArrayList<AgentScorePair> tmpEnsemble = new ArrayList<AgentScorePair>();
		for(Agent agent : _lastPredictions.keySet()) {
			double score = 0;
			ArrayList<Double> rightList = rightMap.get(agent);
			ArrayList<Double> probWeightList = probWeightMap.get(agent);
			for(int  i = 0; i < _numPastRounds && i < rightList.size(); i++) {
				score += rightList.get(rightList.size()-1-i)*_numRightWeight + probWeightList.get(probWeightList.size()-1-i)*_probWeightWeight;
			}

			tmpEnsemble.add(new AgentScorePair(agent,score));
		}
		Collections.sort(tmpEnsemble);
		_weights = new HashMap<Agent, Double>();
		for(int i = 0; i < _numOracles && i < tmpEnsemble.size(); i++) {
			_weights.put(tmpEnsemble.get(i).getOracle(),tmpEnsemble.get(i).getScore());
		}
		
		/*
		 * Get new predictions
		 */
		for(Agent agent : _lastPredictions.keySet()) {
			_lastPredictions.put(agent, agent.getActionDistribution());
		}
	}

	public int isMax(double weight, double[] predictions) {
		for(int i = 0; i < predictions.length; i++) {
			if(predictions[i] > weight) {
				return 0;
			}
		}
		return 1;
	}

	private class AgentScorePair implements Comparable<AgentScorePair> {
		Agent _agent;
		double _score;

		AgentScorePair(Agent agent, double score) {
			_agent = agent;
			_score = score;
		}

		public double getScore() {
			return _score;
		}

		public Agent getOracle() {
			return _agent;
		}

		public int compareTo(AgentScorePair agentScorePair) {
			double thisScore = this._score;
			double otherScore = agentScorePair.getScore();
			if(thisScore < otherScore) {
				return 1;
			}
			if(otherScore < thisScore) {
				return -1;
			}
			else {
				return 0;
			}
		}
	}

	public EnsembleOracle getCopy() {
		return new EnsembleOracle(_g,_seed,_options);
	}

	public String toString() {
		return "Ensemble Oracle(numRightWeight: " + _numRightWeight + ", probWeightWeight: " + _probWeightWeight + ", numPastRounds: " + _numPastRounds + ", numOracles: " + _numOracles + ")";
	}

	@Override
	public Agent getCopy(Game g, long seed, String options) {
		return new EnsembleOracle(g,seed,options);
	}
}