package human;
import game.Game;

import java.awt.BasicStroke;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Stroke;
import java.util.Vector;

import lemonade.Action;
import lemonade.IslandLogic;


/**
 * A view for a human to play the game. The meat of the
 * state of the data is in the HumanModel to which this points.
 * @author maz
 *
 */
public class HumanDisplay extends Canvas implements HumanView{
  /**
   * 
   */
  private static final long serialVersionUID = 1L;
  private Game game;
  private IslandLogic logic;
  
  private HumanModelData model;
  
  
  
  private Color[] colors;
  private Color[] fadedColors;
  
  private static final Color[] DEFAULT_COLORS = new Color[]{new Color(255,0,255),new Color(0,196,0),Color.red,Color.blue,new Color(0,234,0)};
  
  private static final int MAX_UTILITY = 12;
  /**
   * The ocean is a pale blue-green
   */
  private static final Color OCEAN_COLOR = new Color(0,0,255);

  /**
   * The beach is a light yellow
   */
  private static final Color BEACH_COLOR = new Color(255,255,96);
  
  /**
   * The jungle (island interior) is a dark green
   */
  private static final Color JUNGLE_COLOR = new Color(0,128,0);
  
  static final Color VACANT_SPOT_COLOR = Color.black;
  
  
  public HumanDisplay(Game g){
    assert(g.getNumPlayers()==3);
    
    model = new HumanModelData(g.getNumPlayers());
    this.game=g;
    this.logic = new IslandLogic(g.getNumActions());
    colors = new Color[g.getNumPlayers()];
    System.arraycopy(DEFAULT_COLORS, 0, colors, 0, colors.length);
    fadedColors = new Color[g.getNumPlayers()];
    for(int i=0;i<colors.length;i++){
      int red = colors[i].getRed();
      int green = colors[i].getGreen();
      int blue = colors[i].getBlue();
      int fadedRed = red*2/3;
      int fadedGreen = green*2/3;
      int fadedBlue = blue*2/3;
      fadedColors[i]=new Color(fadedRed,fadedGreen,fadedBlue);
    }
  }
  
  /**
   * The ideal size of the canvas.
   */
  private static final Dimension size = new Dimension(640,480);
  
  @Override
  public Dimension getMaximumSize() {
    return size;
  }



  @Override
  public Dimension getMinimumSize() {
    return size;
  }



  @Override
  public Dimension getPreferredSize() {
    return size;
  }

  private int width;
  private int height;
  private int minDim;
  private int radius;
  private int positionRadius;
  private int centerx;
  private int centery;

  private void evaluateResize(){
    width = getWidth();
    height = getHeight();
    minDim = Math.min(width, height);
    radius = (minDim*2)/5;
    positionRadius = radius/10;
    centerx = width/2;
    centery = height/2;
  }
  
  
  /**
   * Get the location of the point (in terms of the positions in the game). -1 indicates no circle was 
   * clicked on.
   * @param p the query point
   * @return the position (0 to islandSize-1), or -1 if a position was not clicked.
   */
  public int getPosition(Point p){
    int x = p.x;
    int y = p.y;
    int positionRadiusSquared = positionRadius*positionRadius;
    Point center = new Point();
    for(int i=0;i<game.getNumActions();i++){
      getPositionPoint(i,center);
      int deltaX = (x-center.x);
      int deltaY = (y-center.y);
      int distanceSquared = (deltaX*deltaX)+(deltaY*deltaY);
      if (distanceSquared<=positionRadiusSquared){
        return i;
      }
    }
    return -1;
  }

  private void getPositionPoint(int position, Point p){
    double angle = Math.PI*2*position/game.getNumActions();
    p.x = centerx+(int)(radius * Math.sin(angle));
    p.y = centery+(int)(radius * Math.cos(angle));
    
  }

  /**
   * Repaints the display. By the contract of HumanModelListener, since this is called after
   * handleChange, this is not allowed to obtain any locks.
   */
  public void paint(Graphics g){
    evaluateResize();
    int pieceRadius = positionRadius/2;

    int pieceDiameter = pieceRadius*2;
    int positionDiameter = positionRadius*2;
    
    int beachOffset = ((positionRadius*5)/4);

    
    g.setColor(HumanDisplay.OCEAN_COLOR);
    
    g.fillRect(0,0,width,height);

    int borderRespected = height/24;

    int beachRadius = radius + beachOffset;
    int beachDiameter = beachRadius*2;
    int beachTLX = centerx-beachRadius;
    int beachTLY = centery-beachRadius;
    
    g.setColor(HumanDisplay.BEACH_COLOR);
    g.fillOval(beachTLX, beachTLY, beachDiameter, beachDiameter);
    
    int jungleRadius = radius - beachOffset;
    int jungleDiameter = jungleRadius*2;
    int jungleTLX = centerx-jungleRadius;
    int jungleTLY = centery-jungleRadius;
    g.setColor(HumanDisplay.JUNGLE_COLOR);
    g.fillOval(jungleTLX, jungleTLY, jungleDiameter, jungleDiameter);
    
    
    Point p = new Point();
    for(int i=0;i<game.getNumActions();i++){
      g.setColor(HumanDisplay.VACANT_SPOT_COLOR);
      this.getPositionPoint(i, p);
      int cornerPosX = p.x-positionRadius;
      int cornerPosY = p.y-positionRadius;
      g.fillOval(cornerPosX, cornerPosY, positionDiameter, positionDiameter);
      //g.setColor(Color.white);
      //g.drawString(""+i, p.x, p.y);
    }
    for(int i=0;i<model.positions.length;i++){
      if (model.isVisible(i)){
        if (model.isFaded(i)){
          g.setColor(fadedColors[i]);
        } else {
          g.setColor(colors[i]);
        }

        this.getPositionPoint(model.positions[i],p);
        double angle = Math.PI*2*i/game.getNumPlayers();
        int pieceX = p.x+(int)(pieceRadius * Math.sin(angle));
        int pieceY = p.y+(int)(pieceRadius * Math.cos(angle));
        int cornerPieceX = pieceX-pieceRadius;
        int cornerPieceY = pieceY-pieceRadius;
        g.fillOval(cornerPieceX, cornerPieceY, pieceDiameter, pieceDiameter);
      }
    }
    if (model.allVisible() && model.utilitiesVisible){
      Vector<Action> sortedPositions = logic.getPositionsSorted(model.positions);
      Graphics2D g2 = (Graphics2D)g;

      Stroke s = g2.getStroke();

      for(int i=0;i<sortedPositions.size();i++){
        Action a = sortedPositions.get(i);
        int piecesOnSpotBefore=0;
        for(int j=0;j<a.player;j++){
          if (model.positions[j]==a.position){
            piecesOnSpotBefore++;
          }
        }
        int posDegrees = 360*a.position/game.getNumActions()-90;
        int arcClockwise = -180*logic.getDistanceRight(sortedPositions, i)/game.getNumActions();
        int arcCounterClockwise = 180*logic.getDistanceLeft(sortedPositions, i)/game.getNumActions();
        int numOnTop = logic.numberOfPlayersOnTop(sortedPositions, i);
        float lineWidth = ((float)pieceRadius)/numOnTop;
        int lineRadius = beachRadius-1+(int)(lineWidth*(0.5+piecesOnSpotBefore));
        g2.setColor(colors[a.player]);
        BasicStroke s2 = new BasicStroke(lineWidth,BasicStroke.CAP_SQUARE,BasicStroke.JOIN_ROUND);
        g2.setStroke(s2);
        g.drawArc(centerx-lineRadius, centery-lineRadius, lineRadius*2, lineRadius*2, posDegrees,arcClockwise);
        g.drawArc(centerx-lineRadius, centery-lineRadius, lineRadius*2, lineRadius*2, posDegrees,arcCounterClockwise);

      }
      
      int legendHeight = height/6;
      int utilScale = legendHeight/MAX_UTILITY;
      legendHeight = utilScale* MAX_UTILITY;
      int legendWidth = width/6;
      int legendx = width-(borderRespected+legendWidth);
      // The lower part of the legend.
      int legendy = height-borderRespected;
      int legendBorder = 5;
      int strokeSize = legendWidth*2/(3*game.getNumPlayers());
      int spacing = (legendWidth-(strokeSize+(legendBorder*2)))/(game.getNumPlayers()-1);
      int[] utility = game.getUtility().getUtility(model.positions);
      
      g2.setStroke(s);
      g2.setColor(Color.white);
      
      for(int i=0;i<=12;i+=2){
        g2.drawLine(legendx, legendy-(utilScale*i), legendx+legendWidth, legendy-(utilScale*i));
      }
      
      
      g2.setStroke(new BasicStroke((float)strokeSize,BasicStroke.CAP_BUTT,BasicStroke.JOIN_MITER));
      for(int i=0;i<utility.length;i++){
        int xCoord = legendx+(spacing*i)+(strokeSize/2)+legendBorder;
        int lineTop = legendy-(utilScale*utility[i]);
        int lineBottom = legendy;
        //System.out.println(lineTop);
        g2.setColor(colors[i]);
        g2.drawLine(xCoord,lineTop,xCoord,lineBottom);
      }
      g2.setStroke(new BasicStroke((float)3,BasicStroke.CAP_BUTT,BasicStroke.JOIN_MITER));
      g2.setColor(Color.white);
      //g2.drawLine(legendx, legendy, legendx, legendy-legendHeight);
      g2.drawRect(legendx, legendy-legendHeight, legendWidth, legendHeight);
      g2.setStroke(s);
      
    }
    String modelMessage = model.message;
    g.setColor(Color.white);
    if (modelMessage!=null){
      int stringWidth = g.getFontMetrics().charsWidth(modelMessage.toCharArray(), 0, modelMessage.length());
      g.drawString(model.message,centerx-(stringWidth/2),centery);
    }
    FontMetrics fm =g.getFontMetrics();
    String []names = {"Purple(You)","Green","Red"};
    int maxNameWidth = 0;
    for(int i=0;i<names.length;i++){
      int nameWidth = fm.charsWidth(names[i].toCharArray(), 0, names[i].length());
      maxNameWidth = Math.max(nameWidth, maxNameWidth);
      int y = borderRespected+(i*g.getFontMetrics().getHeight())+g.getFontMetrics().getAscent();
      int x = borderRespected;
      g.drawString(names[i],x,y);
      
      
    }
    String biggestScore=""+(20*12);
    int scoreLegendWidth = fm.charsWidth(biggestScore.toCharArray(), 0, biggestScore.length());
    for(int i=0;i<model.scores.length;i++){
      String scoreString = ""+model.scores[i];
      int stringWidth = fm.charsWidth(scoreString.toCharArray(), 0, scoreString.length());
      int y = borderRespected+(i*g.getFontMetrics().getHeight())+g.getFontMetrics().getAscent();
      int x = borderRespected+(scoreLegendWidth+maxNameWidth-stringWidth);
      g.drawString(scoreString,x,y);
    }
    if (model.currentGameIndex!=null){
    int gameIndexX=borderRespected;
    int gameIndexY=height-borderRespected;
    
    g.drawString(""+model.currentGameIndex, gameIndexX, gameIndexY);
    }
  }



  @Override
  public void handleChange(HumanModel model) {
    model.getModel(this.model);
    repaint();
  }
}
