package human;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import util.Array;


import game.Game;
import game.Strategy;

/**
 * Serves as a human strategy to play against computer strategies.
 * The Model-View-Controller pattern is approximated, with this class
 * being the controller, HumanDisplay being the view, and HumanModel being
 * the model.
 * 
 * @author maz
 *
 */
public class LocalHumanStrategy implements MouseListener,Strategy{
  private HumanView display;
  private HumanModelData modelData;
  private HumanModel model;
  private Game g;
  
  private JFrame frame;
  private int actionChosen;
  private long timeOfObservation;
  
  private int actualPlayerIndex;
  
  JTextArea area;
  JScrollPane pane;
  public Component getComponent(){
    return frame;
  }
  
  
  public LocalHumanStrategy(Game g, long seed, String features){
    this(g,new HumanDisplay(g));
    HumanDisplay hdisplay = (HumanDisplay)display;
    hdisplay.addMouseListener(this);
    //area = new JTable(new Object[0][3],new String[]{"Purple","Green","Red"});
    
    area = new JTextArea(22,22);
    area.setEditable(false);
    
    area.setBackground(Color.BLUE);
    area.setForeground(Color.WHITE);
    
    area.setFont(new Font("Monospaced",Font.PLAIN,12));
    area.append("Purple Green   Red\n");
    pane = new JScrollPane(area);
    pane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
    //pane.setPreferredSize(new Dimension(200,600));
    frame = new JFrame("Lemonade Stand Game");
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frame.setLayout(new BorderLayout());
    frame.add(pane,BorderLayout.EAST);
    frame.add(hdisplay,BorderLayout.CENTER);
    frame.pack();
    frame.validate();
    frame.setLocationRelativeTo(null);
    frame.setVisible(true);
  }
  

  /**
   * A direct construction of HumanStrategy.
   * Afterwards, HumanView must (in some way) add HumanStrategy
   * as a MouseListener.
   * @param g
   * @param display
   */
  public LocalHumanStrategy(Game g,HumanView display){
    this.g = g;
    actionChosen = -1;
    modelData = new HumanModelData(g.getNumPlayers());
    modelData.setBeginGame();
    model = new HumanModel(g.getNumPlayers());
    this.display = display;
    model.addListener(display);
    display.handleChange(model);
  }
  
  @Override
  public void mouseClicked(MouseEvent e) {
    int location = display.getPosition(e.getPoint());
    if (location!=-1){
      clickOnLocation(location);
    }
  }
  
  /**
   * Public permission is granted to access this function
   * because it allows one to simulate the clicks.
   * @param location the "position" (i.e. action) of the click
   */
  public void clickOnLocation(int location){
    if (chooseAction(location)){
      try{
      model.lockAndGetModel(modelData);
      modelData.setPiece(actualPlayerIndex, location);
      modelData.utilitiesVisible=false;
      modelData.message = "Action locked in...";
      model.setModelAndUnlock(modelData);
      } catch (InterruptedException ie){
        
      }
    }
  }
  
  
  private synchronized boolean chooseAction(int location) {
    if (actionChosen==-1){
      actionChosen = location;
      return true;
    }
    return false;
  }
  
  public synchronized void unlockAction() {
    timeOfObservation = System.currentTimeMillis();
    actionChosen = -1;
  }
  
  public synchronized void lockAction() {
    actionChosen = -2;
  }
  
  

  private synchronized int queryAction() {
    int result = -1;
    if (actionChosen!=-1){
      result=actionChosen;
      actionChosen = -2;
    }
    return result;
  }
  
  
  @Override
  public void mouseEntered(MouseEvent e) {
  }
  @Override
  public void mouseExited(MouseEvent e) {
  }
  @Override
  public void mousePressed(MouseEvent e) {
  }
  @Override
  public void mouseReleased(MouseEvent e) {
  }
  
  @Override
  public int getAction() {
    try{
    while(true){
      int possible=queryAction();
      if (possible!=-1){
        return possible;
      }
      
      long timeAfterObservation = System.currentTimeMillis()-timeOfObservation;
      if (timeAfterObservation>5000 && modelData.utilitiesVisible){
        model.lockAndGetModel(modelData);
        modelData.utilitiesVisible=false;
        for(int i=0;i<modelData.positions.length;i++){
          if (i!=actualPlayerIndex){
            modelData.setFaded(i, true);
          }
        }
        modelData.message = "Choose an action...";
        model.setModelAndUnlock(modelData);
      } 
      Thread.sleep(100);
    }
    } catch (InterruptedException ie){
      return g.getNumActions();
    } 
  }
  
  public static String rightJustifyArray(int[] actions){
    String result = "";
    for(int action:actions){
      result += String.format("%6d", action);
    }
    return result;
  }
  @Override
  public void observeOutcome(int[] actions) {
    assert(g.areLegalActions(actions));
    
    area.append(rightJustifyArray(actions)+"\n");
    area.revalidate();
    Dimension d = area.getSize();
    Rectangle r = new Rectangle(0,d.height,0,0);
    area.scrollRectToVisible(r);
    area.repaint();
    try{
    Array.swap(actions,0,actualPlayerIndex);
    model.lockAndGetModel(modelData);
    
    for(int i=0;i<actions.length;i++){
      modelData.setPiece(i,actions[i]);
    }
    modelData.setFadedAll(false);
    modelData.utilitiesVisible=true;
    int[] utilityChange = g.getUtility().getUtility(actions);
    int pointsWon = utilityChange[actualPlayerIndex];
    
    modelData.message = "Points won:"+pointsWon;
    assert(modelData.scores.length==utilityChange.length);
    
    for(int i=0;i<actions.length;i++){
      modelData.scores[i]+=utilityChange[i];
    }
    model.setModelAndUnlock(modelData);
    } catch (InterruptedException ie){
      ie.printStackTrace();
    }
    timeOfObservation = System.currentTimeMillis();
    unlockAction();
  }

  public HumanModel getModel() {
    return model;
  }

  public void dispose() {
    frame.dispose();
  }

  public void setActualPlayerIndex(int actualPlayerIndex) {
    this.actualPlayerIndex = actualPlayerIndex;
  }
  
}
